#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for LED."""

from __future__ import with_statement

import libpal.palcmd as palcmd
from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import (ExecError, InvalidArgError, NotFoundError)
from pal_config import (SET_LED_CMD, I2CSET_CMD)
import pal_config as config
from pal_info import (Results, ReqTypes)
import os

### XML Tag Name ###
LED_NAME = "led"
POWER_NAME = "power"
STATUS_NAME = "status"
HDD_NAME = "hdd"
BRIGHTNESS_NAME = "brightness"
ANY_CHILD_NAME = "*"

### XML Attribute ###
COLOR_ATTR = "color"
HDD_NUM_ATTR = "hddnum"

### XML Attribute value ###
HDD1_ATTR_VAL = "1"
HDD2_ATTR_VAL = "2"
HDD3_ATTR_VAL = "3"
HDD4_ATTR_VAL = "4"
HDD5_ATTR_VAL = "5"
HDD6_ATTR_VAL = "6"
RED = "red"
GREEN = "green"
BLUE = "blue"
ORANGE = "orange"
NON_COLER = "NOCOLOR"

### XML Element ###
ON_EL = "on"
BLINK_EL = "blink"
OFF_EL = "off"
BRIGHTNESS_LEVEL1_EL = "25"
BRIGHTNESS_LEVEL2_EL = "50"
BRIGHTNESS_LEVEL3_EL = "75"
BRIGHTNESS_LEVEL4_EL = "100"

LED_ON = "on"
LED_BLINK = "blink"
LED_OFF = "off"

VAR = "%s"

CTRL_STATUS_LED_ARG = [SET_LED_CMD, "STATUS", VAR, VAR]
CTRL_STATUS_LED_VAR_INDEX_LIST = [2, 3]

SET_LED_BRIGHTNESS_CMD_1 = \
    [I2CSET_CMD, "-y", VAR, "0x32", "0x00", "0x40"]

SET_LED_BRIGHTNESS_CMD_1_VAR_INDEX_LIST = [2]

SET_LED_BRIGHTNESS_CMD_2 = \
    [I2CSET_CMD, "-y", VAR, "0x32", "0x01", "0x0c"]

SET_LED_BRIGHTNESS_CMD_2_VAR_INDEX_LIST = [2]

SET_LED_BRIGHTNESS_CMD_3 = \
    [I2CSET_CMD, "-y", VAR, "0x32", "0x03", VAR]

SET_LED_BRIGHTNESS_CMD_3_VAR_INDEX_LIST = [2, 5]

SET_LED_BRIGHTNESS_CMD_4 = \
    [I2CSET_CMD, "-y", VAR, "0x32", "0x08", "0x39"]

SET_LED_BRIGHTNESS_CMD_4_VAR_INDEX_LIST = [2]

SET_LED_BRIGHTNESS_CMD_5 = \
    [I2CSET_CMD, "-y", VAR, "0x52", "0x20", VAR]

SET_LED_BRIGHTNESS_CMD_5_VAR_INDEX_LIST = [2, 5]

STATUS_LED_MODE_MAP = {ON_EL: "1",
                       BLINK_EL: "2",
                       OFF_EL: "0"}

STATUS_LED_DEV_MAP = {RED: "led_error",
                      GREEN: "led_good"}


HDD_LED_MODE_MAP = {ON_EL: "1",
                    BLINK_EL: "2",
                    OFF_EL: "0"}

HDD_LED_DEV_MAP = {BLUE: {"0": "hdd0_link",
                          "1": "hdd1_link",
                          "2": "hdd2_link",
                          "3": "hdd3_link"},
                   RED: {"0": "hdd0_err",
                         "1": "hdd1_err",
                         "2": "hdd2_err",
                         "3": "hdd3_err"}}

### Command option value ###
ON_CMD_OP = "ON"
BLINK_CMD_OP = "BLINK"
OFF_CMD_OP = "OFF"
RED_CMD_OP = "RED"
GREEN_CMD_OP = "GREEN"
ORANGE_CMD_OP = "ORANGE"

class PalLed(palcmd.PalCmd):
    """Pal Led command implementation class."""

    #
    # Sub functions
    #

    def __check_correct_root_node(self, entry):
        """Check whether root node is correctly.
        """

        node_name = entry.get().get_node_name()
        if node_name != LED_NAME:
            raise InvalidArgError(node_name)

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __get_value_by_key(self, target_key, dict):
        """This method is for utility.
           Find key of "dict" and return corresponding key.
        """

        for key, value in dict.iteritems():
            if target_key == key:
                return value
        raise InvalidArgError("Invalid key:" + str(target_key))

    def __get_i2c_bus(self):
        """Get i2c bus."""

        I2C_NAME_PATH = "/sys/bus/i2c/devices/i2c-%u/name"

        for i in range(0, 16):
            if not os.path.exists(I2C_NAME_PATH % i):
                continue

            line = open(I2C_NAME_PATH% i, "r").readline()
            if line.startswith("SMBus"):
                return "%u" % i

        raise NotFoundError

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError("Command list size is bad.. (__execute_cmd)")

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __transact_status_led_node(self, status_led_node):
        """Transact status node.
        """

        try:
            color_attr_val = status_led_node.get_attr_value(COLOR_ATTR)
            color_op = self.__get_value_by_key(color_attr_val,
                                               {RED: RED_CMD_OP,
                                                GREEN: GREEN_CMD_OP,
                                                ORANGE: ORANGE_CMD_OP})
        except NotFoundError:
            color_op = GREEN_CMD_OP

        mode_op = \
            self.__get_value_by_key(status_led_node.get_text(),
                                    {ON_EL: ON_CMD_OP,
                                     BLINK_EL: BLINK_CMD_OP,
                                     OFF_EL: OFF_CMD_OP})

        self.__execute_cmd(True, True, NULL, NULL,
                           CTRL_STATUS_LED_ARG,
                           CTRL_STATUS_LED_VAR_INDEX_LIST,
                           [color_op, mode_op])
        return

    def __get_hddnum_from_attr(self, target_key):
        """Get HDD number(integer) by specified key"""

        return self.__get_value_by_key(target_key,
                                       {HDD1_ATTR_VAL: 0,
                                        HDD2_ATTR_VAL: 1,
                                        HDD3_ATTR_VAL: 2,
                                        HDD4_ATTR_VAL: 3,
                                        HDD5_ATTR_VAL: 4,
                                        HDD6_ATTR_VAL: 5})

    def __get_cmd_hdd_map_cmd_opt(self, hdd_light_on_status_dict):

        map_opt_value = 0xf000
        for hdd_num, color_dict in hdd_light_on_status_dict.iteritems():

            if color_dict[RED] << hdd_num:
                map_opt_value += 0x01 << hdd_num * 2

            if color_dict[BLUE] << hdd_num:
                map_opt_value += 0x02 << hdd_num * 2
        return hex(int(map_opt_value))

    def __hold_hdd_led_color(self, hdd_num, color):

        REQ_COLOR_TO_HODL_COLOR = \
            {BLUE: config.HOLD_STATE_BLUE,
             RED: config.HOLD_STATE_RED,
             NON_COLER: config.HOLD_STATE_OFF}

        with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "w") as fh:
            fh.write(REQ_COLOR_TO_HODL_COLOR[color])

    def __get_current_hdd_led_color(self, hdd_num):

        try:
            with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "r") as fh:
                return fh.read().rstrip()
        except:
            return OFF_EL

    def __transact_hdd_led_status_node_list(self, hdd_led_status_node_list):
        """Transact ctrl hdd status led node.
           <hdd>
             <state color="red" hddnum="1" >on</state>
        """

        for status_node in hdd_led_status_node_list:

            # HDD number
            hddnum_attr_val = status_node.get_attr_value(HDD_NUM_ATTR)
            hdd_num = self.__get_hddnum_from_attr(hddnum_attr_val)

            # LED mode
            mode = status_node.get_text()
            led_mode = HDD_LED_MODE_MAP[mode]

            # LED color
            try:
                tgt_color = status_node.get_attr_value(COLOR_ATTR)
                if OFF_EL == mode:
                    tgt_color = BLUE
            except NotFoundError:
                tgt_color = BLUE

            for color in  HDD_LED_DEV_MAP.keys():
                led_dev_path = \
                    os.path.join(config.DEV_DIR,
                                 HDD_LED_DEV_MAP[color][str(hdd_num)])
                if color == tgt_color:
                    with open(led_dev_path, "w") as fh:
                        fh.write(led_mode)
                else:
                    # 指定された色以外は消灯
                    with open(led_dev_path, "w") as fh:
                        fh.write(HDD_LED_MODE_MAP[OFF_EL])
        return

    def __transact_hdd_brightness_node(self, hdd_led_brightness_node):
        """Transact HDD LED brightness node."""

        BRIGHTNESS_LEVEL_MAP= \
            {BRIGHTNESS_LEVEL1_EL: "0x00",
             BRIGHTNESS_LEVEL2_EL: "0x10",
             BRIGHTNESS_LEVEL3_EL: "0x35",
             BRIGHTNESS_LEVEL4_EL: "0xFF",}

        i2c_bus_num = self.__get_i2c_bus()

        brightness = hdd_led_brightness_node.get_text()

        self.__execute_cmd(True, True, NULL, NULL,
                           SET_LED_BRIGHTNESS_CMD_1,
                           SET_LED_BRIGHTNESS_CMD_1_VAR_INDEX_LIST,
                           [i2c_bus_num])

        self.__execute_cmd(True, True, NULL, NULL,
                           SET_LED_BRIGHTNESS_CMD_2,
                           SET_LED_BRIGHTNESS_CMD_2_VAR_INDEX_LIST,
                           [i2c_bus_num])
                           
        self.__execute_cmd(True, True, NULL, NULL,
                           SET_LED_BRIGHTNESS_CMD_3,
                           SET_LED_BRIGHTNESS_CMD_3_VAR_INDEX_LIST,
                           [i2c_bus_num, BRIGHTNESS_LEVEL_MAP[brightness]])
        
        self.__execute_cmd(True, True, NULL, NULL,
                           SET_LED_BRIGHTNESS_CMD_4,
                           SET_LED_BRIGHTNESS_CMD_4_VAR_INDEX_LIST,
                           [i2c_bus_num])

        self.__execute_cmd(True, True, NULL, NULL,
                           SET_LED_BRIGHTNESS_CMD_5,
                           SET_LED_BRIGHTNESS_CMD_5_VAR_INDEX_LIST,
                           [i2c_bus_num, BRIGHTNESS_LEVEL_MAP[brightness]])

    def __transact_hdd_led_node(self, hdd_led_node):
        """Transact hdd led node."""

        STATUS_XPATH = "statuslist/status"

        # HDD STATUS LED
        hdd_led_status_node_list = hdd_led_node.get_node_list(STATUS_XPATH)
        if len(hdd_led_status_node_list) > 0:
            self.__transact_hdd_led_status_node_list(hdd_led_status_node_list)

        return

    #
    # Request handlers
    #

    def __control_handler(self, entry):
        """Handle "control" request.
        """

        self.__check_correct_root_node(entry)

        target_led_node_list = entry.get().get_node_list(ANY_CHILD_NAME)
        for target_led_node in target_led_node_list:
            node_name = target_led_node.get_node_name()

            if STATUS_NAME == node_name:
                self.__transact_status_led_node(target_led_node)
                continue

            if HDD_NAME == node_name:
                self.__transact_hdd_led_node(target_led_node)
                continue

            if BRIGHTNESS_NAME == node_name:
                self.__transact_hdd_brightness_node(target_led_node)
                continue

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.CONTROL, self.__control_handler)

#
# Main
#

if __name__ == "__main__":
    PalLed(option=palcmd.PalCmdOption()).handle_request()
