#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

from __future__ import with_statement

import os
import sys
import shutil
import errno
from ionas.request import Request, Entry, EntryList
from iobase.xmldata import XmlNode, ANY_CHILD_PATH
from iobase.cmdexec import CmdExec, NULL
from iobase.exception import NotFoundError
from fal_info import (ReqTypes, Results,
    FAL_MESSAGES_PATH, FAL_VALIDATE_PATH,
    FAL_SERVICECTRL_PATH, FAL_PORTS_PATH,
    FAL_VOLCTRL_PATH, FAL_EVENTS_PATH)
from product_conf import BIN_DIR, XML_CONF_DIR, STOR_IF_TO_VOL
from product_conf.devinfo import USB_PORT_NO_LIST

ADD_LANDISKMANUAL_CMD = "evh_landiskmanual_share_add.py"
ADD_LANDISKMANUAL_CMD_PATH = \
    "/mnt/data/libexec/evh/" + ADD_LANDISKMANUAL_CMD
MOUNT_EVENT_XML = \
    "<events><eventinfo category='volume' action='mount' priority='75'>" + \
      "<volume><name>sataraid1</name></volume></eventinfo></events>"


def _request(fal, data):
    return Request(fal).request(data)


def request(cmd, type=None, data=None):
    if type:
        if not data:
            return _request(cmd, Entry(type)).\
                get_entrylist().get_entry()
        if isinstance(data, str):
            return _request(cmd, Entry(type, XmlNode(data))).\
                get_entrylist().get_entry()
        if isinstance(data, XmlNode):
            return _request(cmd, Entry(type, data)).\
                get_entrylist().get_entry()
    else:
        if isinstance(data, EntryList):
            return  _request(cmd, data).\
                get_entrylist()
        if (isinstance(data, Entry) or
            isinstance(data, XmlNode)):
            return  _request(cmd, data).\
                get_entrylist().get_entry()


def _service_ctrl(service, type):
    request(FAL_SERVICECTRL_PATH, type, """<servicectrl>
            <servicelist>
                <service><name>%s</name></service>
            </servicelist>
        </servicectrl>""" % service)


def service_start(service, is_check_mounted=False):
    if is_check_mounted:
        if not is_mounted_internal_volume():
            return
    _service_ctrl(service, ReqTypes.START)


def service_stop(service):
    _service_ctrl(service, ReqTypes.STOP)


def service_restart(service):
    _service_ctrl(service, ReqTypes.SUSPEND)
    _service_ctrl(service, ReqTypes.RESUME)


def _get_tip_node(node, fullpath):
    for path in fullpath.split("/"):
        try:
            node.get_node(path)
        except NotFoundError:
            try:
                node.append_child(path)
            except ValueError:
                # xpathである場合に"["以前を抜き出す
                path = path.split("[")[0]
                node.append_child(path)
        finally:
            node = node.get_node_list(path)[-1]
    return node


def _get_target_node(org_msg_xml, path, id_name, id_val):
    xpath = "%s[%s='%s']" % (path, id_name, id_val)
    try:
        target_node = org_msg_xml.get_node(xpath)
        target_node.delete_children(ANY_CHILD_PATH)
    except NotFoundError:
        target_node = _get_tip_node(org_msg_xml, xpath)

    return target_node


def _xml_modify(org_msg_xml, new_msg_node, path, identifier):
    new_node_list = new_msg_node.get_node_list(path)

    for new_node in new_node_list:
        target_node = _get_target_node(
            org_msg_xml, path, identifier, new_node.get_value(identifier))
        target_node.update(new_node)


def msg_xml_modify(new_msg_node):
    BUILD_MSGINFO_PATH = BIN_DIR + "/build_msginfo"
    MESSAGES_XML_PATH = XML_CONF_DIR + "/messages.xml"

    # fal_msginfoのファイルパス
    # pycの場合のためcをrstripしている
    import fal_msginfo
    FAL_MSGINFO_PATH = fal_msginfo.__file__.rstrip("c")

    res_entry = request(FAL_MESSAGES_PATH, ReqTypes.LOAD)
    if res_entry.get_result() != Results.OK:
        raise Exception

    org_msg_xml = res_entry.get()

    _xml_modify(org_msg_xml, new_msg_node,
        "categorylist/category", "code")
    _xml_modify(org_msg_xml, new_msg_node,
        "subcategorylist/subcategory", "code")
    _xml_modify(org_msg_xml, new_msg_node,
        "messageinfolist/messageinfo", "code")

    res_entry = request(FAL_MESSAGES_PATH, ReqTypes.SAVE, org_msg_xml)
    if res_entry.get_result() != Results.OK:
        raise Exception

    CmdExec([BUILD_MSGINFO_PATH, MESSAGES_XML_PATH, FAL_MSGINFO_PATH],
        stdout=NULL, stderr=NULL).execute()


def valid_xml_modify(new_msg_node):
    BUILD_VALIDINFO_PATH = BIN_DIR + "/build_validinfo"
    VALIDATES_XML_PATH = XML_CONF_DIR + "/validates.xml"

    # fal_msginfoのファイルパス
    # pycの場合のためcをrstripしている
    import fal_validinfo
    FAL_VALIDINFO_PATH = fal_validinfo.__file__.rstrip("c")

    res_entry = request(FAL_VALIDATE_PATH, ReqTypes.LOAD)
    if res_entry.get_result() != Results.OK:
        raise Exception

    org_msg_xml = res_entry.get()

    _xml_modify(org_msg_xml, new_msg_node,
        "validatelist/validate", "name")
    _xml_modify(org_msg_xml, new_msg_node,
        "altlist/alt", "@name")

    res_entry = request(FAL_VALIDATE_PATH, ReqTypes.SAVE, org_msg_xml)
    if res_entry.get_result() != Results.OK:
        raise Exception

    CmdExec([BUILD_VALIDINFO_PATH, VALIDATES_XML_PATH, FAL_VALIDINFO_PATH],
        stdout=NULL, stderr=NULL).execute()


def msg_xml_modify_with_file(pkg_name):
    FILE_NAME = "messages.xml"

    file_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        pkg_name, FILE_NAME)

    with open(file_path) as file:
        xml_node = XmlNode(file.read())

    msg_xml_modify(xml_node)


def valid_xml_modify_with_file(pkg_name):
    FILE_NAME = "validates.xml"

    file_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        pkg_name, FILE_NAME)

    with open(file_path) as file:
        xml_node = XmlNode(file.read())

    valid_xml_modify(xml_node)


def remove_files(file_list):
    for file in file_list:
        if os.path.isdir(file):
            shutil.rmtree(file)
        elif os.path.isfile(file):
            os.unlink(file)


def _enable_usb_conf(usb_conf_path, port, enable=True):
    SUFFIX_LIST = ["ro", "rw"]
    ORIG_NAME = "usb%s.%s.real"
    LINK_NAME = "usb%s.%s"

    for suffix in SUFFIX_LIST:
        orig_path = os.path.join(usb_conf_path, ORIG_NAME % (port, suffix))
        link_path = os.path.join(usb_conf_path, LINK_NAME % (port, suffix))

        if enable:
            if not os.path.exists(orig_path):
                return
            try:
                os.symlink(orig_path, link_path)
            except (IOError, OSError), ins:
                if ins.errno != errno.EEXIST:
                    raise
        else:
            try:
                os.unlink(link_path)
            except (IOError, OSError), ins:
                if ins.errno != errno.ENOENT:
                    raise


def ctrl_usb_conf(usb_conf_path):
    USB_SHARE_ENABLE = \
        "usblist/usb[@port='%s']/modelist/mode[text()='share']/@enable"

    port_conf = request(FAL_PORTS_PATH, ReqTypes.LOAD).get()

    for port in USB_PORT_NO_LIST:
        is_enable_usb_share = port_conf.get_value(USB_SHARE_ENABLE % port)
        if is_enable_usb_share == "true":
            _enable_usb_conf(usb_conf_path + port, port, True)
        else:
            _enable_usb_conf(usb_conf_path + port, port, False)


def update_conf(fal):
    request(fal, ReqTypes.UPDATE)


def update_webui():
    # 下記ファイルを触ることで
    # djangoのwebページ認識が更新される

    WEBUI_UPDATE_FILE = "/usr/local/share/www/webui/webui/wsgi.py"
    os.utime(WEBUI_UPDATE_FILE, None)


def set_default_conf(fal, fpath):
    conf_node = XmlNode(open(fpath).read())
    request(fal, ReqTypes.SAVE, conf_node)


def request_conf_entrylist(conf_cmd, conf_xml_node, tg_list_xpath, req_type,
                           pack_xpath="", tg_node_xpath=""):
    entry_list = EntryList()

    def is_tg_node(node):
        if not tg_node_xpath:
            return True

        try:
            node.get_node(tg_node_xpath)
            return True
        except NotFoundError:
            return False

    for node in conf_xml_node.get_node_list(tg_list_xpath):
        if not is_tg_node(node):
            continue

        if pack_xpath:
            pack_xpath_sep_list = pack_xpath.split("/")
            body = XmlNode(pack_xpath_sep_list[0])
            body_tip_node = _get_tip_node(body,
                                          "/".join(pack_xpath_sep_list[1:]))
            body_tip_node.append_child(node)
        else:
            body = node

        entry_list.append_entry(Entry(req_type, body))

    if not entry_list.get_entry_list():
        return

    request(conf_cmd, data=entry_list)


def update_real_conf(conf_cmd, conf_fal, tg_list_xpath,
                           pack_xpath="", tg_node_xpath=""):

    conf_xml_node = request(conf_fal, ReqTypes.LOAD).get()

    request_conf_entrylist(conf_cmd, conf_xml_node, tg_list_xpath,
                           ReqTypes.EVENT, pack_xpath, tg_node_xpath)


def is_mounted_volume(vol_name):

    result_entry = request(
        FAL_VOLCTRL_PATH,
        ReqTypes.QUERY,
        XmlNode(
            """<volctrl>
                <volume>
                    <name type="volume">%s</name>
                </volume>
            </volctrl>""" % vol_name))

    if result_entry.get_result() == Results.DEVICE_NOT_FOUND:
        return False

    if result_entry.get().get_value("volume/status") == "mounted":
        return True

    return False


def is_mounted_volume_list(vol_list):

    for vol in vol_list:
        if not is_mounted_volume(vol):
            return False
    return True


def is_mounted_internal_volume():

    return is_mounted_volume_list(STOR_IF_TO_VOL["internal"].values())


def set_package_manual(cmd):

    def regist_after_event():
        request(FAL_EVENTS_PATH,
            data=Entry(ReqTypes.REGISTER, MOUNT_EVENT_XML,
                file_name=cmd)).get_result()

    result = request(cmd,
        ReqTypes.EVENT, MOUNT_EVENT_XML).get_result()

    if result != Results.OK:
        regist_after_event()


class BaseXmlConfModifier():

    FAL = ""
    MODIFY_INFO_LIST = []

    def install(self, pre_remove=True):
        # landisk adminからの設定等によってすでに同じnodeが
        # 存在する可能性あるので事前に削除
        if pre_remove:
            self.remove()

        config = request(self.FAL, ReqTypes.LOAD).get()
        for modify_info in self.MODIFY_INFO_LIST:
            for par_node in config.get_node_list(modify_info.PAR_XPATH):
                for install_node in modify_info.INSTALL_NODE_LIST:
                    par_node.append_child(install_node.get_clone())
        request(self.FAL, ReqTypes.SAVE, config)

    def remove(self):
        config = request(self.FAL, ReqTypes.LOAD).get()
        for modify_info in self.MODIFY_INFO_LIST:
            for remove_xpath in modify_info.REMOVE_XPATH_LIST:
                try:
                    config.delete_nodes(modify_info.PAR_XPATH, remove_xpath)
                except NotFoundError:
                    pass
        request(self.FAL, ReqTypes.SAVE, config)

    def upgrade(self):
        config = request(self.FAL, ReqTypes.LOAD).get()
        for modify_info in self.MODIFY_INFO_LIST:
            for par_node in config.get_node_list(modify_info.PAR_XPATH):
                for upgrade_node in modify_info.UPGRADE_NODE_LIST:
                    try:
                        par_node.get_node(upgrade_node.get_node_name())
                    except NotFoundError:
                        par_node.append_child(upgrade_node.get_clone())
        request(self.FAL, ReqTypes.SAVE, config)


if __name__ == "__main__":
    ins = BaseXmlConfModifier()
    if sys.argv[1] == "install":
        ins.install()
    elif sys.argv[1] == "remove":
        ins.remove()
