#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""Utility module for Udev Event Handler."""


from __future__ import with_statement

import re
import os
import sys
import time

import iobase.xmldata as xmldata
import iobase.fileut as fileut
import ionas.request as request
from iobase.cmdexec import (CmdExec, NULL)

from product_conf import *
from product_conf.devinfo import (HDD_DEV_LIST,)

import fal_info as info
from fal_info import (ReqTypes, Results)


UDEV_TEMP_DIR = os.path.join(TEMP_DIR, "udev")
LOCK_FILE_BASE = "udev_"

STATEMGR_QUERY_XML = '<statemgr><statelist><state category="system"/>' + \
                        '</statelist></statemgr>'
STATE_PATH = "./statelist/state"

EVENTINFO_NAME = "eventinfo"
CATEGORY_ATTR_NAME = "category"
ACTION_ATTR_NAME = "action"
PREVIOUS_ATTR_NAME = "previous"

EVENTLOG_BODY_TMPL = "<eventlog><log><code>%s</code></log></eventlog>"
LOG_NAME = "log"
ARGLIST_NAME = "arglist"
ARG_NAME = "arg"
NAME_NAME = "name"
NAME_ATTR_NAME = "name"

EVENT_ACTION_PLUG = "plug"
EVENT_ACTION_UNPLUG = "unplug"

RUNNING_STATE = "running"
BOOTING_STATE = "booting"
CANCEL_STATE_LIST = ("updating", "initalizing", "shutdown")
STATE_CHECK_PERIOD = 5

UDEV_SETTLE_OPT = "settle"

ACTION_ENV = "ACTION"
LINK_NAME_ENV = "LINK_NAME"

DEVICE_EVENT_BODY_TMPL = "<device><name>%s</name></device>"

DEVICE_UNKNOWN= "unknown"

ACTION_ADD = "add"
ACTION_REMOVE = "remove"
ACTION_CHANGE = "change"

LED_BLINK = "blink"
LED_ON = "on"
LED_BODY_TMPL = '<led><status color="green">%s</status></led>'

BUZZER_BODY_TMPL = '<buzzer><melody type="%s"/></buzzer>'

TRUE = xmldata.XmlConfNode.TRUE_TEXT

DEV_LINK_DIR = os.path.join(STATUS_TEMP_DIR, "dev_link")
DEV_FILE_DIR = os.path.join(STATUS_TEMP_DIR, "dev_file")

DEV_NULL = "/dev/null"


class _NotMatch:
    """Exception class for representing that "re.match()" return None."""

    pass


def _dev_to_port_info(dev_name, pattern):
    """Return tuple of port I/F name and number if "dev_name" match with
       "pattern".
    """

    match = re.match(pattern, dev_name.strip())
    if match == None:
        raise _NotMatch

    return (match.group(1), match.group(2))


def dev_to_port_info(dev_name):
    """Return tuple of port I/F name and number."""

    dev_name = os.path.basename(dev_name)
    try:
        return _dev_to_port_info(dev_name, "(.+)d([0-9]+).*")
    except:
        pass

    try:
        return _dev_to_port_info(dev_name, "(.+)ups([0-9]+).*")
    except:
        pass

    try:
        return _dev_to_port_info(dev_name, "(.+)lp([0-9]+).*")
    except:
        pass

    return ("", "")


def request_entry(fal_cmd, entry):
    """Send "entry" to "fal_cmd" and return the entry in response data."""

    return request.Request(fal_cmd).request(entry).get_entry()


def get_system_state():
    """Return the state of "system"."""

    try:
        entry = request.Entry(ReqTypes.QUERY, STATEMGR_QUERY_XML)
        response_entry = request_entry(info.FAL_STATEMGR_PATH, entry)

        if response_entry.get_result() != Results.OK:
            return BOOTING_STATE

        return response_entry.get().get_node(STATE_PATH).get_text()
    except:
        return BOOTING_STATE


def wait_system_ready():
    """Wait the system state to become ready."""

    while True:
        if get_system_state() == RUNNING_STATE:
            return
        time.sleep(STATE_CHECK_PERIOD)


def log_eventlog(code, args={}):
    """Log to eventlog."""

    entry = request.Entry(ReqTypes.ADD, EVENTLOG_BODY_TMPL % (code,))
    if len(args):
        arglist = xmldata.XmlNode(ARGLIST_NAME)
        for arg in args:
            arglist.append_child(ARG_NAME, args[arg], {NAME_ATTR_NAME: arg})
        entry.get().get_node(LOG_NAME).append_child(arglist)

    request_entry(info.FAL_EVENTLOG_PATH, entry)


def get_device_event_body(device_name):
    """Return event body for device event."""

    return xmldata.XmlNode(DEVICE_EVENT_BODY_TMPL % (device_name,))


def dispatch_event(category, action, body=None, previous=False):
    """Dispatch event."""

    entry = request.Entry(ReqTypes.EVENT,
                info.FAL_EVENTS_REQ_BASE_XML[ReqTypes.EVENT])

    attrs = {CATEGORY_ATTR_NAME: category, ACTION_ATTR_NAME: action}
    if previous:
        attrs[PREVIOUS_ATTR_NAME] = TRUE

    entry.get().get_node(EVENTINFO_NAME).set_attrs(attrs)
    if body != None:
        if isinstance(body, (str, unicode)):
            body = xmldata.XmlNode(body)
        entry.get().get_node(EVENTINFO_NAME).append_child(body)

    return request_entry(info.FAL_EVENTS_PATH, entry)


def try_background_exec():
    """Try execute own in background."""

    pid = os.fork()
    if pid < 0:
        # Force execution
        reurn

    if pid > 0:
        sys.exit(0)

    # Redirect stdin, stdout and stderr to null device.
    null_fd = os.open(DEV_NULL, os.O_RDWR)
    for fd in (0, 1, 2):
        os.close(fd)
        os.dup2(null_fd, fd)


def wait_udev_settle():
    """Wait for the udev event queue to finish."""

    CmdExec((UDEVADM_CMD, UDEV_SETTLE_OPT)).execute()


def get_device_name():
    """Return target device name (normally symbolic link name)."""

    if LINK_NAME_ENV in os.environ:
        return os.environ[LINK_NAME_ENV]
    return DEVICE_UNKNOWN


def is_action_add():
    """Return action is "add" or not."""

    try:
        return os.environ[ACTION_ENV] == ACTION_ADD
    except:
        return False


def is_action_remove():
    """Return action is "remove" or not."""

    try:
        return os.environ[ACTION_ENV] == ACTION_REMOVE
    except:
        return False


def is_action_change():
    """Return action is "change" or not."""

    try:
        return os.environ[ACTION_ENV] == ACTION_CHANGE
    except:
        return False


def lock_udev_info(device_name):
    """Try locking the use of "device_name" by creating file and
       return result - 0 : success  1 : fail.
    """

    with fileut.FLock(LOCK_DIR, device_name, exclusive=True):
        udev_info_file = os.path.join(UDEV_TEMP_DIR, device_name)
        if os.path.exists(udev_info_file):
            return 1

        if not os.path.exists(UDEV_TEMP_DIR):
            os.makedirs(UDEV_TEMP_DIR)

        with open(udev_info_file, "w"):
            pass
        return 0


def unlock_udev_info(device_name):
    """Unlock the use of "device_name" by deleting file."""

    with fileut.FLock(LOCK_DIR, device_name, exclusive=True):
        udev_info_file = os.path.join(UDEV_TEMP_DIR, device_name)
        if os.path.exists(udev_info_file):
            os.remove(udev_info_file)


def is_udev_info_locked(device_name):
    """Return if the use of "device_name" is locked or not."""

    with fileut.FLock(LOCK_DIR, device_name, exclusive=False):
        udev_info_file = os.path.join(UDEV_TEMP_DIR, device_name)
        if os.path.exists(udev_info_file):
            return True
        return False


def try_udev_info_lock_or_unlock(device_name):
    """Try locking or unlocking the use of "device_name" according to
       udev action and return result - 0 : success  1 : fail.
    """

    if device_name == DEVICE_UNKNOWN:
        return 1

    if is_action_add() or is_action_change():
        if is_udev_info_locked(device_name):
            return 1
        return lock_udev_info(device_name)

    if is_action_remove():
        if not is_udev_info_locked(device_name):
            return 1
        unlock_udev_info(device_name)
        return 0

    return 1


def set_status_led(led_condition):
    """Set the status LED to "led_condition"."""

    request_entry(info.FAL_LED_PATH,
            request.Entry(ReqTypes.CONTROL, LED_BODY_TMPL % (led_condition,)))


def set_buzzer(melody):
    """Set the buzzer to "melody"."""

    request_entry(info.FAL_BUZZER_PATH,
            request.Entry(ReqTypes.PLAY, BUZZER_BODY_TMPL % (melody,)))
